#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wrap/atomics.h"
#include "wrap/synch.h"
#include "wrap/thread.h"

/**
 * En faktura som ska skickas till en kund.
 */
struct invoice {
    // E-postadress som fakturan ska skickas till.
    const char *to;

    // Belopp som ska betalas in.
    int amount;

    // OCR-nummer som kunden ska ange för att vi ska veta vem som betalade.
    int ocr_code;
};

// Skapa en faktura.
struct invoice *invoice_create(const char *to, int amount, int ocr_code) {
    struct invoice *result = malloc(sizeof(struct invoice));
    result->to = to;
    result->amount = amount;
    result->ocr_code = ocr_code;
    return result;
}

// Städa upp en skapad faktura.
void invoice_destroy(struct invoice *invoice) {
    free(invoice);
}

// Skicka en faktura (till en fil i nuläget).
void invoice_send(struct invoice *invoice, FILE *to) {
    fprintf(to, "Faktura till %s:\n Belopp: %d kr\n OCR: %d\n\n",
            invoice->to, invoice->amount, invoice->ocr_code);
}


// Maximalt antal fakturor som kan ligga i kön samtidigt.
#define QUEUE_SIZE 10

/**
 * En anslutning till en e-postserver.
 *
 * För att miska belastningen ser vi till att vi bara har en anslutning igång
 * till varje e-postserver. Däremot vill vi kunna ha flera anslutningar till
 * olika servrar öppna samtidigt för att kunna skicka ut så många fakturor som
 * möjligt åt gången.
 *
 * I det här exempelprogrammet simulerar vi skickandet av e-postmeddelanden
 * genom att skriva till filer i stället för att ansluta till en server över
 * Internet.
 *
 * Anslutningen innehåller en kö av fakturor som den ska se till att skicka ut i
 * någon ordning. När anslutningen skapas, skapas också en tråd som är ansvarig
 * för att faktiskt skicka meddelandena (denna tråd kör funktionen
 * 'connection_thread').
 */
struct connection {
    // Adress till den server vi är ansluten.
    const char *server;

    // Den fil vi har öppnat för att skriva fakturor till. I den riktiga
    // implementationen blir detta en nätverksanslutning i stället.
    FILE *output;

    // Array av fakturaobjekt vi ska skicka. Vi ser till att tomma element
    // innehåller NULL.
    struct invoice *to_send[QUEUE_SIZE];

    // Lås för "to_send".
    struct lock to_send_lock;

    // Antal använda platser i "to_send"
    struct semaphore to_send_count;

    // Ska vi avsluta tråden som hör till den här anslutningen?
    bool terminate_thread;
};

// Hur många trådar är klara nu?
struct semaphore thread_terminated;

// Funktion som körs för varje connection-instans som skapas (i en egen
// tråd). Har ansvar för att skicka meddelandena som läggs i kön. Vi antar att
// funktionen bara anropas inifrån "connection_create".
static void connection_thread(void *data) {
    struct connection *connection = data;

    do {
        bool queue_empty = true;

        // Skicka alla meddelanden vi kan hitta.
        for (int i = 0; i < QUEUE_SIZE; i++) {
            lock_acquire(&connection->to_send_lock);
            struct invoice *invoice = connection->to_send[i];
            lock_release(&connection->to_send_lock);
            if (invoice) {
                connection->to_send[i] = NULL;
                sema_down(&connection->to_send_count);

                invoice_send(invoice, connection->output);
                invoice_destroy(invoice);
                queue_empty = false;
            }
        }

        // Om vi hittade något, titta igen eftersom det kan finnas mer att
        // skicka.
        if (!queue_empty)
            continue;

        // Om vi inte ska avsluta, fortsätt loopa.
    } while (!connection->terminate_thread);

    // Meddela att vi är klara.
    sema_up(&thread_terminated);
}

// Skapa en anslutning till en server, och starta en tråd som skickar de
// fakturor som läggs till sedan.
struct connection *connection_create(const char *server) {
    struct lock lock;
    lock_init(&lock);
    lock_acquire(&lock);

    struct connection *result = malloc(sizeof(struct connection));
    result->server = server;
    // Notera: 'fopen' öppnar en fil. Det går bra att anropa fopen från flera
    // trådar samtidigt. Det går även att skriva till olika filer från olika
    // trådar samtidigt.
    result->output = fopen(server, "w");
    result->terminate_thread = false;
    for (int i = 0; i < QUEUE_SIZE; i++) {
        result->to_send[i] = NULL;
    }
    lock_init(&result->to_send_lock);
    sema_init(&result->to_send_count, QUEUE_SIZE);

    // Skapa en tråd som är ansvarig för att skicka meddelanden.
    thread_new(&connection_thread, result);

    lock_release(&lock);

    return result;
}

// Stäng anslutningen och se till att frigöra alla resurser som hör till
// (dvs. stänga filer, stoppa tråden). Funktionen väntar på att alla fakturor
// som har lagts till har skickats ut.
void connection_destroy(struct connection *connection) {
    // Be tråden att avsluta sig, och vänta på att den har gjort det.
    connection->terminate_thread = true;
    sema_down(&thread_terminated);

    // Nu när tråden är klar vet vi att "to_send" är tom.
    fclose(connection->output);
    free(connection);
}

// Lägg till en faktura till en anslutning. Vi antar att fakturan ska skickas
// till den adress anslutningen hanterar. Skulle det vara så att "to_send" redan
// är full väntar funktionen på att det finns plats. Funktionen kommer se till
// att frigöra "invoice". Fakturorna skickas inte nödvändigtvis ut i samma
// ordning som de lades till i kön.
void connection_send(struct connection *to, struct invoice *invoice) {
    lock_acquire(&to->to_send_lock);
    int i = 0;
    while (true) {
        if (to->to_send[i] == NULL) {
            sema_up(&to->to_send_count);
            to->to_send[i] = invoice;
            break;
        }

        // Stega till nästa element.
        i = (i + 1) % QUEUE_SIZE;
    }
    lock_release(&to->to_send_lock);
}


/**
 * Bonusuppgift.
 *
 * Om även nedastående del av koden synkroniseras ges upp till 2 bonuspoäng.
 */


// Hitta delen efter "@" i en e-postadress.
const char *find_server(const char *original) {
    for (int i = 0; original[i]; i++)
        if (original[i] == '@')
            return original + i + 1;

    // Inget "@", returnera allt helt enkelt.
    return original;
}

#define MAX_CONNECTIONS 10

// Alla anslutningar vi har öppna just nu.
struct connection *open_connections[MAX_CONNECTIONS];

// Lås för öppna anslutningar.
struct lock open_lock;

// Initiera globala data. Anropas en gång i början av programmet.
void init() {
    lock_init(&open_lock);
    sema_init(&thread_terminated, 0);
    for (int i = 0; i < MAX_CONNECTIONS; i++)
        open_connections[i] = NULL;
}

// Skicka ett meddelande. Öppnar en ny anslutning om det behövs.
void send(struct invoice *invoice) {
    const char *server = find_server(invoice->to);

    for (int i = 0; i < MAX_CONNECTIONS; i++) {
        lock_acquire(&open_lock);
        struct connection *c = open_connections[i];
        if (c == NULL) {
            // Vi hittade NULL. Då fanns inte anslutningen, vi skapar en ny.
            c = connection_create(server);
            open_connections[i] = c;
        }

        // Är det rätt anslutning?
        if (strcmp(c->server, server) == 0) {
            connection_send(c, invoice);
            lock_release(&open_lock);
            return;
        }
        lock_release(&open_lock);
    }
}

// Stäng alla anslutningar. Vi antar att detta görs en gång i slutet av programmet.
void close_all(void) {
    lock_acquire(&open_lock);
    for (int i = 0; i < MAX_CONNECTIONS; i++) {
        if (open_connections[i]) {
            connection_destroy(open_connections[i]);
            open_connections[i] = NULL;
        }
    }
    lock_release(&open_lock);
}


/**
 * Huvudprogram.
 *
 * Endast för att visa hur vi kan köra programmet. Vid rättning kommer
 * eventuella modifikationer härifrån och nedåt att ignoreras.
 */

int main() {
    init();

    send(invoice_create("filip.stromback@liu.se", 134, 512431));
    send(invoice_create("klas.arvidsson@liu.se", 110, 512731));
    send(invoice_create("daniel.thoren@liu.se", 99, 512438));
    send(invoice_create("andan123@student.liu.se", 20, 512345));
    send(invoice_create("filst914@student.liu.se", 15, 561662));

    close_all();
    return 0;
}
